from __future__ import absolute_import

import sys

from .client import get_client

try:
    # Django >= 1.10
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    # Not required for Django <= 1.9, see:
    # https://docs.djangoproject.com/en/1.10/topics/http/middleware/#upgrading-pre-django-1-10-style-middleware
    MiddlewareMixin = object


class AukletMiddleware(MiddlewareMixin):
    def process_exception(self, request, exception):
        exc_type, _, traceback = sys.exc_info()
        client = get_client()
        client.produce_event(exc_type, traceback)


class WSGIAukletMiddleware(object):
    """
    A WSGI middleware which will attempt to capture any
    uncaught exceptions and send them to Auklet.
    """

    def __init__(self, application):
        self.application = application

    def __call__(self, environ, start_response):
        """Call the application can catch exceptions."""
        app = None
        try:
            app = self.application(environ, start_response)
            for item in app:
                yield item
        # Catch any exception
        except Exception:
            self.handle_exception(environ)

        if hasattr(app, 'close'):
            app.close()

    def handle_exception(self, environ=None):
        exc_type, _, traceback = sys.exc_info()
        client = get_client()
        client.produce_event(exc_type, traceback)



import msgpack

from time import time
from uuid import uuid4

from django.conf import settings

from auklet.errors import AukletConfigurationError
from auklet.broker import MQTTClient
from auklet.stats import Event, SystemMetrics, FilenameCaches
from auklet.utils import get_agent_version, get_device_ip, get_mac, \
                         create_dir

try:
    # For Python 3.0 and later
    from urllib.error import HTTPError, URLError
    from urllib.request import Request, urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen, Request, HTTPError, URLError


_client = None


def get_client():
    """
    Get an Auklet Client
    """
    global _client
    if _client is None:
        _client = DjangoClient()
    return _client


class DjangoClient(object):
    def __init__(self):
        auklet_config = settings.AUKLET_CONFIG
        self.apikey = auklet_config.get("api_key", None)
        self.app_id = auklet_config.get("application", None)
        self.release = auklet_config.get("release", None)
        self.version = auklet_config.get("version", None)
        self.org_id = auklet_config.get("organization", None)
        self.broker_url = auklet_config.get("broker", "mq.feeds.auklet.io")
        self.port = auklet_config.get("port", 8883)
        self.base_url = auklet_config.get("base_url", "https://api.auklet.io/")

        if self.apikey is None:
            raise AukletConfigurationError(
                "Please set api_key in AUKLET_CONFIG settings")
        if self.app_id is None:
            raise AukletConfigurationError(
                "Please set application in AUKLET_CONFIG settings")
        if self.org_id is None:
            raise AukletConfigurationError(
                "Please set organization in AUKLET_CONFIG settings")
        self.auklet_dir = create_dir()
        self.mac_hash = get_mac()
        self.device_ip = get_device_ip()
        self.agent_version = get_agent_version()
        self.broker = MQTTClient(self.broker_url, self.port, self.app_id,
                                 self.org_id, self.apikey, self.base_url,
                                 self.auklet_dir)
        self.file_cache = FilenameCaches()

    def build_event_data(self, type, traceback):
        event = Event(type, traceback, self.file_cache)
        event_dict = dict(event)
        event_dict['application'] = self.app_id
        event_dict['publicIP'] = get_device_ip()
        event_dict['id'] = str(uuid4())
        event_dict['timestamp'] = int(round(time() * 1000))
        event_dict['systemMetrics'] = dict(SystemMetrics())
        event_dict['macAddressHash'] = self.mac_hash
        event_dict['release'] = self.release
        event_dict['version'] = self.version
        event_dict['agentVersion'] = get_agent_version()
        event_dict['device'] = None
        return event_dict

    def build_msgpack_event_data(self, type, traceback):
        event_data = self.build_event_data(type, traceback)
        return msgpack.packb(event_data, use_bin_type=False)

    def produce_event(self, type, traceback):
        self.broker.produce(self.build_msgpack_event_data(type, traceback))


def init_client():
    global client
    client = DjangoClient()



from __future__ import absolute_import, division, unicode_literals

import inspect

try:
    import psutil
except ImportError:  # pragma: no cover
    # Some platforms that applications could be running on require specific
    # installation of psutil which we cannot configure currently
    psutil = None

__all__ = ['Event', 'SystemMetrics', 'FilenameCaches']


class Event(object):
    __slots__ = ['trace', 'exc_type', 'line_num']

    def __init__(self, exc_type, tb, tree):
        self.exc_type = exc_type.__name__
        self.line_num = tb.tb_lineno
        self._build_traceback(tb, tree)

    def __iter__(self):
        yield "stackTrace", self.trace
        yield "excType", self.exc_type

    def _convert_locals_to_string(self, local_vars):
        for key in local_vars:
            if type(local_vars[key]) != str and type(local_vars[key]) != int:
                local_vars[key] = str(local_vars[key])
        return local_vars

    def _build_traceback(self, trace, tree):
        tb = []
        while trace:
            frame = trace.tb_frame
            path = tree.get_filename(frame.f_code, frame)
            tb.append({"functionName": frame.f_code.co_name,
                       "filePath": path,
                       "lineNumber": frame.f_lineno,
                       "locals":
                           self._convert_locals_to_string(frame.f_locals)})
            trace = trace.tb_next
        self.trace = tb


class FilenameCaches(object):
    cached_filenames = {}

    def get_filename(self, code, frame):
        key = code.co_code
        file_name = self.cached_filenames.get(code.co_code, None)
        if file_name is None:
            try:
                file_name = inspect.getsourcefile(frame) or \
                            inspect.getfile(frame)
            except (TypeError, AttributeError):
                # These functions will fail if the frame is of a
                # built-in module, class or function
                return None
            self.cached_filenames[key] = file_name
        return file_name


class SystemMetrics(object):
    cpu_usage = 0.0
    mem_usage = 0.0
    inbound_network = 0
    outbound_network = 0
    prev_inbound = 0
    prev_outbound = 0

    def __init__(self):
        if psutil is not None:
            self.cpu_usage = psutil.cpu_percent(interval=1)
            self.mem_usage = psutil.virtual_memory().percent
            network = psutil.net_io_counters()
            self.prev_inbound = network.bytes_recv
            self.prev_outbound = network.bytes_sent

    def __iter__(self):
        yield "cpuUsage", self.cpu_usage
        yield "memoryUsage", self.mem_usage
        yield "inboundNetwork", self.inbound_network
        yield "outboundNetwork", self.outbound_network

    def update_network(self, interval):
        if psutil is not None:
            network = psutil.net_io_counters()
            self.inbound_network = (network.bytes_recv -
                                    self.prev_inbound) / interval
            self.outbound_network = (network.bytes_sent -
                                     self.prev_outbound) / interval
            self.prev_inbound = network.bytes_recv
            self.prev_outbound = network.bytes_sent



from __future__ import absolute_import

import os
import ssl
import logging
import paho.mqtt.client as mqtt

from auklet.utils import build_url, create_file

try:
    # For Python 3.0 and later
    from urllib.error import HTTPError, URLError
    from urllib.request import Request, urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen, Request, HTTPError, URLError

__all__ = ["MQTTClient"]


class MQTTClient(object):
    producer = None
    brokers = "mq.feeds.auklet.io"
    client = None
    username = None
    password = None
    port = 8883

    def __init__(self, broker_url, port, app_id, org_id, apikey, base_url,
                 auklet_dir):
        self.brokers = broker_url
        self.port = int(port)
        self.org_id = org_id
        self.app_id = app_id
        self.apikey = apikey
        self.base_url = base_url
        self.auklet_dir = auklet_dir
        self.create_producer()
        topic_suffix = "{}/{}".format(
            self.org_id, self.app_id)
        self.producer_types = {
            "monitoring": "django/profiler/{}".format(topic_suffix),
            "event": "django/events/{}".format(topic_suffix),
        }

    def _get_certs(self):
        if not os.path.isfile("{}/ca.pem".format(self.auklet_dir)):
            url = Request(
                build_url(self.base_url, "private/devices/certificates/"),
                headers={"Authorization": "JWT {}".format(self.apikey)})
            try:
                try:
                    res = urlopen(url)
                except HTTPError as e:
                    # Allow for accessing redirect w/o including the
                    # Authorization token.
                    res = urlopen(e.geturl())
            except URLError:
                return False
            filename = "{}/ca.pem".format(self.auklet_dir)
            create_file(filename)
            f = open(filename, "wb")
            f.write(res.read())
            return True
        return False

    def on_disconnect(self, client, userdata, rc):
        if rc != 0:
            logging.debug("Unexpected disconnection from MQTT")

    def create_producer(self):
        if self._get_certs():
            self.producer = mqtt.Client(client_id=self.app_id,
                                        protocol=mqtt.MQTTv311,
                                        transport="ssl")
            self.producer.username_pw_set(
                username=self.app_id,
                password=self.apikey)
            self.producer.enable_logger()
            context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            context.verify_mode = ssl.CERT_REQUIRED
            context.load_verify_locations(capath="{}/".format(self.auklet_dir))
            context.options &= ~ssl.OP_NO_SSLv3
            self.producer.tls_set_context()
            self.producer.on_disconnect = self.on_disconnect
            self.producer.connect_async(self.brokers, self.port)
            self.producer.loop_start()

    def produce(self, data, data_type="monitoring"):
        self.producer.publish(self.producer_types[data_type], payload=data)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django.apps import AppConfig


class AukletConfig(AppConfig):
    name = "auklet.client"
    label = "auklet_client"
    verbose_name = "Auklet"

    def ready(self):
        from .client import init_client
        init_client()



import os
import sys
import uuid
import hashlib
import requests
import tempfile

from auklet.__about__ import __version__ as auklet_version

__all__ = ['open_auklet_url', 'create_file', 'clear_file', 'build_url',
           'get_mac', 'get_device_ip', 'get_agent_version',
           'post_auklet_url', 'create_dir', 'b', 'u']


def open_auklet_url(url, apikey):
    try:
        res = requests.get(url, headers={"Authorization": "JWT {}".format(apikey)})
    except Exception:
        return None
    return res


def post_auklet_url(url, apikey, data):
    try:
        res = requests.post(
            url,
            json=data,
            headers={"Authorization": "JWT {}".format(apikey),
                     "Content-Type": "application/json"})
    except requests.HTTPError:
        return None
    return res.json()


def create_dir(dir_name=".auklet"):
    dirs = [os.getcwd(), os.path.expanduser("~")]
    for directory in dirs:
        if os.access(directory, os.W_OK):
            if not os.path.exists(".auklet"):
                os.mkdir(dir_name)
            return directory
    return tempfile.gettempdir()


def create_file(filename):
    open(filename, "a").close()


def clear_file(filename):
    open(filename, "w").close()


def build_url(base_url, extension):
    return '%s%s' % (base_url, extension)


def get_mac():
    mac_num = hex(uuid.getnode()).replace('0x', '').upper()
    mac = '-'.join(mac_num[i: i + 2] for i in range(0, 11, 2))
    return hashlib.md5(b(mac)).hexdigest()


def get_device_ip():
    try:
        res = requests.get("https://api.ipify.org")
        return u(res.content)
    except (requests.RequestException, Exception):
        # TODO log to kafka if the ip service fails for any reason
        return None


def get_agent_version():
    return auklet_version


if sys.version_info < (3,):
    # Python 2 and 3 String Compatibility
    def b(x):
        return x

    def u(x):
        return x
else:
    # https://pythonhosted.org/six/#binary-and-text-data
    import codecs

    def b(x):
        # Produces a unicode string to encoded bytes
        return codecs.utf_8_encode(x)[0]

    def u(x):
        # Produces a byte string from a unicode object
        return codecs.utf_8_decode(x)[0]
